@@ -39,11 +39,13 @@
 #include "window.h"
 #include <gcu/objprops.h>
 #include <gcu/loader.h>
-#include <goffice/math/go-rangefunc.h>
+#ifndef GOFFICE_HAS_GLOBAL_HEADER
+#   include <goffice/math/go-rangefunc.h>
+#   include <gtk/gtk.h>
+#   include <gio/gio.h>
+#endif
 #include <unistd.h>
 #include <libgen.h>
-#include <gtk/gtk.h>
-#include <gio/gio.h>
 #include <glib/gi18n-lib.h>
 #include <openbabel/mol.h>
 #include <openbabel/obconversion.h>
@@ -191,11 +193,13 @@ void Document::SetFileName (string const &Name, const gchar* mime_type)
 	list<string>::iterator cur, end = exts.end ();
 	for (cur = exts.begin (); cur != end; cur++)
 		if (*cur == ext) {
-			m_label = g_strndup (m_filename + i, j - i);
+			char *buf = g_strndup (m_filename + i, j - i);
+			m_label = g_uri_unescape_string (buf, NULL);
+			g_free (buf);
 			break;
 		}
 	if (!m_label)
-		m_label = g_strdup (m_filename + i);
+		m_label = g_uri_unescape_string (m_filename + i, NULL);
 }
 
 void Document::BuildBondList (list<Bond*>& BondList, Object const *obj) const
@@ -426,7 +430,7 @@ void Document::DoPrint (GtkPrintOperation *print, GtkPrintContext *context) cons
 	WidgetData* pData = (WidgetData*) g_object_get_data (G_OBJECT (widget), "data");
 	ArtDRect rect;
 	pData->GetObjectBounds (this, &rect);
-	double scale = 1.;
+	double scale = .75;
 	x = rect.x0;
 	y = rect.y0;
 	w = rect.x1 - rect.x0;
@@ -439,10 +443,6 @@ void Document::DoPrint (GtkPrintOperation *print, GtkPrintContext *context) cons
 		break;
 	case GCU_PRINT_SCALE_FIXED:
 		scale *= Printable::GetScale ();
-		x *= scale;
-		y *= scale;
-		w *= scale;
-		h *= scale;
 		break;
 	case GCU_PRINT_SCALE_AUTO:
 		if (GetHorizFit () && GetVertFit ())
@@ -451,12 +451,12 @@ void Document::DoPrint (GtkPrintOperation *print, GtkPrintContext *context) cons
 			scale *= width / w;
 		else if (GetVertFit ())
 			scale *= height / h;
-		x *= scale;
-		y *= scale;
-		w *= scale;
-		h *= scale;
 		break;
 	}
+	x *= scale;
+	y *= scale;
+	w *= scale;
+	h *= scale;
 	if (GetHorizCentered ())
 		x -= (width - w) / 2;
 	if (GetVertCentered ())
@@ -866,7 +866,7 @@ void Document::RemoveAtom (Atom* pAtom)
 	Bond* pBond;
 	while ((pBond = (Bond*) pAtom->GetFirstBond (i)))
 	{
-		if (!m_bUndoRedo)
+		if (!m_bUndoRedo && !m_bIsLoading && m_pCurOp)
 			m_pCurOp->AddObject (pBond);
 		RemoveBond (pBond);
 	}
@@ -882,7 +882,7 @@ void Document::RemoveFragment (Fragment* pFragment)
 	Atom* pAtom = pFragment->GetAtom ();
 	Bond* pBond;
 	while ((pBond = (Bond*) pAtom->GetFirstBond (i))) {
-		if (!m_bUndoRedo)
+		if (!m_bUndoRedo && !m_bIsLoading && m_pCurOp)
 			m_pCurOp->AddObject (pBond);
 		RemoveBond (pBond);
 	}
@@ -1296,12 +1296,12 @@ double Document::GetYAlign ()
 
 void Document::SetTheme (Theme *theme)
 {
+	if (m_Theme)
+		m_Theme->RemoveClient (this);
 	if (!theme) {
 		m_Theme = NULL;
 		return;
 	}
-	if (m_Theme)
-		m_Theme->RemoveClient (this);
 	m_Theme = theme;
 	m_Theme->AddClient (this);
 	m_BondLength = theme->GetBondLength ();
