@@ -4,7 +4,7 @@
  * Gnome Chemistry Utils
  * gcu/glview.cc 
  *
- * Copyright (C) 2006-2007 Jean Bréfort <jean.brefort@normalesup.org>
+ * Copyright (C) 2006-2008 Jean Bréfort <jean.brefort@normalesup.org>
  *
  * This program is free software; you can redistribute it and/or 
  * modify it under the terms of the GNU General Public License as 
@@ -26,25 +26,38 @@
 #include <gcu/application.h>
 #include <gcu/gldocument.h>
 #include <gcu/glview.h>
+#include <gcu/macros.h>
+#ifndef GOFFICE_HAS_GLOBAL_HEADER
+#   include <goffice/goffice-features.h>
+#   include <goffice/utils/go-image.h>
+#   include <gio/gio.h>
+#endif
 #include <GL/gl.h>
 #include <GL/glu.h>
-#include <GL/osmesa.h>
 #include <gtk/gtkgl.h>
-#include <libgnomevfs/gnome-vfs-ops.h>
 #include <glib/gi18n-lib.h>
-#include <cairo-svg.h>
 #include <cmath>
 #include <cstring>
 
+#define ROOTDIR "/apps/gchemutils/gl/"
+
 static GdkGLConfig *glconfig = NULL;
 double DefaultPsi = 70.;
 double DefaultTheta = 10.;
 double DefaultPhi = -90.;
+bool OffScreenRendering = false;
 
 using namespace std;
 
 namespace gcu
 {
+#ifdef HAVE_GO_CONF_SYNC
+GOConfNode *GLView::m_ConfNode = NULL;
+#else
+GConfClient *GLView::m_ConfClient = NULL;
+#endif
+guint GLView::m_NotificationId = 0;
+int GLView::nbViews = 0;
 
 // Callbacks
 static bool on_init(GtkWidget *widget, GLView* View) 
@@ -79,25 +92,38 @@ static bool on_pressed(GtkWidget *widget, GdkEventButton *event, GLView* View)
 	return View->OnPressed (event);
 }
 
-// GLView implementation
+#ifdef HAVE_GO_CONF_SYNC
+static void on_config_changed (GOConfNode *node, gchar const *key, gpointer data)
+{
+	if (!strcmp (key, ROOTDIR"off-screen-rendering"))
+		OffScreenRendering = go_conf_get_bool (node, key);
+}
+#else
+static void on_config_changed (GConfClient *client, guint cnxn_id, GConfEntry *entry, gpointer data)
+{
+	if (!strcmp (gconf_entry_get_key (entry),ROOTDIR"off-screen-rendering"))
+		OffScreenRendering = gconf_value_get_bool (gconf_entry_get_value (entry));
+}
+#endif
 
-GLView::GLView (GLDocument* pDoc)
+// GLView implementation
+#define GCU_CONF_DIR_GL "gl"
+
+GLView::GLView (GLDocument* pDoc) throw (std::runtime_error): Printable ()
 {
 	m_bInit = false;
 	m_Doc = pDoc;
 	m_Red = m_Green = m_Blue = 0.;
 	m_Alpha = 1.;
 	m_Angle = 10.;
+	nbViews++;
 	SetRotation (DefaultPsi, DefaultTheta, DefaultPhi);
 /* Create new OpenGL widget. */
 	if (glconfig == NULL)
 	{
 		/* Check if OpenGL is supported. */
 		if (!gdk_gl_query_extension())
-		{
-			g_print("\n*** OpenGL is not supported.\n");
-			exit(1);
-		}
+			throw  runtime_error ("*** OpenGL is not supported.\n");
 	
 		/* Configure OpenGL-capable visual. */
 	
@@ -106,10 +132,20 @@ GLView::GLView (GLDocument* pDoc)
 											GDK_GL_MODE_DEPTH |
 											GDK_GL_MODE_DOUBLE));
 		if (glconfig == NULL)
-		{
-			g_print("*** Cannot find the double-buffered visual.\n");
-			  exit(1);
-		}
+			throw  runtime_error ("*** Cannot find the double-buffered visual.\n");
+#ifdef HAVE_GO_CONF_SYNC
+		m_ConfNode = go_conf_get_node (Application::GetConfDir (), GCU_CONF_DIR_GL);
+#else
+		GError *error = NULL;
+		m_ConfClient = gconf_client_get_default ();
+		gconf_client_add_dir (m_ConfClient, "/apps/gchemutils/gl", GCONF_CLIENT_PRELOAD_ONELEVEL, NULL);
+#endif
+		GCU_GCONF_GET_NO_CHECK ("off-screen-rendering", bool, OffScreenRendering, true)
+#ifdef HAVE_GO_CONF_SYNC
+		m_NotificationId = go_conf_add_monitor (m_ConfNode, "off-screen-rendering", (GOConfMonitorFunc) on_config_changed, NULL);
+#else
+		m_NotificationId = gconf_client_notify_add (m_ConfClient, "/apps/gchemutils/gl", (GConfClientNotifyFunc) on_config_changed, NULL, NULL, NULL);
+#endif
 	}
 	/* create new OpenGL widget */
 	m_pWidget = GTK_WIDGET(gtk_drawing_area_new());
@@ -149,6 +185,20 @@ GLView::GLView (GLDocument* pDoc)
 
 GLView::~GLView ()
 {
+	nbViews--;
+	if (!nbViews) {
+#ifdef HAVE_GO_CONF_SYNC
+		go_conf_remove_monitor (m_NotificationId);
+		go_conf_free_node (m_ConfNode);
+		m_ConfNode = NULL;
+#else
+		gconf_client_notify_remove (m_ConfClient, m_NotificationId);
+		gconf_client_remove_dir (m_ConfClient, "/apps/gchemutils/gl", NULL);
+		g_object_unref (m_ConfClient);
+		m_ConfClient = NULL;
+#endif
+		m_NotificationId = 0;
+	}
 }
 
 void GLView::Init ()
@@ -304,49 +354,18 @@ void GLView::Rotate (gdouble x, gdouble y)
 	m_Phi /= M_PI / 180.;
 }
 
-void GLView::Print (GnomePrintContext *pc, gdouble width, gdouble height)
-{
-	int Width = m_pWidget->allocation.width;
-	int Height = m_pWidget->allocation.height;
-	if (Width > width) {
-		Height = (int) (Height * width / Width);
-		Width = (int) width;
-	}
-	if (Height > height) {
-		Width = (int) (Width * height / Height);
-		Height = (int) height;
-	}
-	double matrix [6] = {Width, 0, 0, Height,
-						(width - Width) / 2, (height - Height )/ 2};
-	int w= (int) (Width * 300. / 72.), h = (int) (Height * 300. / 72.);
-	GdkPixbuf * pixbuf = BuildPixbuf (w, h);
-
-	if (pixbuf) {
-		gnome_print_gsave (pc);
-		gnome_print_concat (pc, matrix);
-		gnome_print_rgbimage (pc, (const guchar*) gdk_pixbuf_get_pixels (pixbuf), w, h, gdk_pixbuf_get_rowstride (pixbuf));
-		gnome_print_grestore (pc);
-		g_object_unref (pixbuf);
-	}
-}
-
 static gboolean do_save_image (const gchar *buf, gsize count, GError **error, gpointer data)
 {
-	GnomeVFSHandle *handle = (GnomeVFSHandle*) data;
-	GnomeVFSFileSize written = 0;
-	GnomeVFSResult res;
+	GOutputStream *output = (GOutputStream *) data;
 	while (count) {
-		res = gnome_vfs_write (handle, buf, count, &written);
-		if (res != GNOME_VFS_OK) {
-			g_set_error (error, g_quark_from_static_string ("gchemutils"), res, gnome_vfs_result_to_string (res));
+		count -= g_output_stream_write (output, buf, count, NULL, error);
+		if (*error)
 			return false;
-		}
-		count -= written;
 	}
 	return true;
 }
 
-void GLView::SaveAsImage (string const &filename, char const *type, map<string, string>& options, unsigned width, unsigned height)
+void GLView::SaveAsImage (string const &filename, char const *type, map<string, string>& options, unsigned width, unsigned height) const
 {
 	if (width == 0 || height == 0)
 		return;
@@ -363,22 +382,22 @@ void GLView::SaveAsImage (string const &filename, char const *type, map<string,
 			keys[j] = (*i).first.c_str ();
 			values[j++] = (*i).second.c_str ();
 		}
-		GnomeVFSHandle *handle = NULL;
-		if (gnome_vfs_create (&handle, filename.c_str (), GNOME_VFS_OPEN_WRITE, true, 0644) == GNOME_VFS_OK) {
-			gdk_pixbuf_save_to_callbackv (pixbuf, do_save_image, handle, type, (char**) keys, (char**) values, &error);
-			if (error) {
-				fprintf (stderr, _("Unable to save image file: %s\n"), error->message);
-				g_error_free (error);
-			}
-			gnome_vfs_close (handle); // hope there will be no error there
+		GFile *file = g_vfs_get_file_for_uri (g_vfs_get_default (), filename.c_str ());
+		GFileOutputStream *output = g_file_create (file, G_FILE_CREATE_NONE, NULL, &error);
+		if (!error)
+			gdk_pixbuf_save_to_callbackv (pixbuf, do_save_image, output, type, (char**) keys, (char**) values, &error);
+		if (error) {
+			fprintf (stderr, _("Unable to save image file: %s\n"), error->message);
+			g_error_free (error);
 		}
+		g_object_unref (file);
 		g_free (keys);
 		g_free (values);
 		g_object_unref (pixbuf);
 	}
 }
 
-GdkPixbuf *GLView::BuildPixbuf (unsigned width, unsigned height)
+GdkPixbuf *GLView::BuildPixbuf (unsigned width, unsigned height) const
 {
 	GdkGLConfig *glconfig = gdk_gl_config_new_by_mode (
 		GdkGLConfigMode (GDK_GL_MODE_RGBA | GDK_GL_MODE_DEPTH));
@@ -386,11 +405,15 @@ GdkPixbuf *GLView::BuildPixbuf (unsigned width, unsigned height)
 	GdkGLPixmap *gl_pixmap = gdk_pixmap_set_gl_capability (pixmap,
 							       glconfig,
 							       NULL );
-	GdkGLDrawable * drawable = gdk_pixmap_get_gl_drawable (pixmap);
-	GdkGLContext * context = gdk_gl_context_new (drawable,
+	GdkGLDrawable *drawable = NULL;
+	GdkGLContext *context = NULL;
+	if (gl_pixmap != NULL) {
+		drawable = gdk_pixmap_get_gl_drawable (pixmap);
+		context = gdk_gl_context_new (drawable,
 						     NULL,
 						     TRUE,
 						     GDK_GL_RGBA_TYPE);
+	}
 	double aspect = (GLfloat) width / height;
 	double x = m_Doc->GetMaxDist (), w, h;
 	if (x == 0)
@@ -403,7 +426,13 @@ GdkPixbuf *GLView::BuildPixbuf (unsigned width, unsigned height)
 		h = w / aspect;
 	}
 	GdkPixbuf *pixbuf = NULL;
-	if (gdk_gl_drawable_gl_begin (drawable, context)) {
+	gdk_error_trap_push ();
+	bool result = OffScreenRendering && gl_pixmap && gdk_gl_drawable_gl_begin (drawable, context);
+	gdk_flush ();
+	if (gdk_error_trap_pop ())
+		result = false;
+	
+	if (result) {
 	    glEnable (GL_LIGHTING);
 		glEnable (GL_LIGHT0);
 		glEnable (GL_DEPTH_TEST);
@@ -445,6 +474,9 @@ GdkPixbuf *GLView::BuildPixbuf (unsigned width, unsigned height)
 		double dxStep, dyStep;
 		unsigned char *tmp, *dest, *src, *dst;
 		unsigned LineWidth, s = sizeof(int);
+		gtk_window_present (GTK_WINDOW (gtk_widget_get_toplevel (m_pWidget))); 
+		while (gtk_events_pending ())
+			gtk_main_iteration ();
 		if (m_pWidget->allocation.width & (s - 1))
 			LineWidth = ((~(s - 1)) & (m_pWidget->allocation.width * 3)) + s;
 		else
@@ -513,13 +545,78 @@ GdkPixbuf *GLView::BuildPixbuf (unsigned width, unsigned height)
 	} else {
 osmesa:
 		g_warning ("Off-screen rendering not supported in this context");
-		// TODO: implement rendering using an exterbal program and osmesa
+		// TODO: implement rendering using an external program and osmesa
 	}
-	gdk_gl_context_destroy (context);
-	gdk_gl_pixmap_destroy (gl_pixmap);
+	if (context)
+		gdk_gl_context_destroy (context);
+	if (gl_pixmap)
+		gdk_gl_pixmap_destroy (gl_pixmap);
 	// destroying pixmap gives a CRITICAL and destroying glconfig leeds to a crash.
-	Update ();
+	const_cast <GLView *> (this)->Update ();
 	return pixbuf;
 }
 
+void GLView::DoPrint (GtkPrintOperation *print, GtkPrintContext *context) const
+{
+	cairo_t *cr;
+	gdouble width, height;
+
+	cr = gtk_print_context_get_cairo_context (context);
+	width = gtk_print_context_get_width (context);
+	height = gtk_print_context_get_height (context);
+	int w, h; // size in points
+	w = m_pWidget->allocation.width;
+	h = m_pWidget->allocation.height;
+	switch (GetScaleType ()) {
+	case GCU_PRINT_SCALE_NONE:
+		break;
+	case GCU_PRINT_SCALE_FIXED:
+		w *= GetScale ();
+		h *= GetScale ();
+		break;
+	case GCU_PRINT_SCALE_AUTO:
+		if (GetHorizFit ())
+			w = width;
+		if (GetVertFit ())
+			h = height;
+		break;
+	}
+	double scale = 300. / 72.;
+	GdkPixbuf *pixbuf = BuildPixbuf (w * scale, h * scale);
+	GOImage *img = go_image_new_from_pixbuf (pixbuf);
+	cairo_pattern_t *cr_pattern = go_image_create_cairo_pattern (img);
+	cairo_matrix_t cr_matrix;
+	double x = 0., y = 0.;
+	if (GetHorizCentered ())
+		x = (width - w) / 2.;
+	if (GetVertCentered ())
+		y = (height - h) / 2.;
+	cairo_matrix_init_scale (&cr_matrix, scale, scale);
+	cairo_matrix_translate (&cr_matrix, -x, -y);
+	cairo_pattern_set_matrix (cr_pattern, &cr_matrix);
+	cairo_rectangle (cr, x, y, w, h);
+	cairo_set_source (cr, cr_pattern);
+	cairo_fill (cr);
+	cairo_pattern_destroy (cr_pattern);
+	g_object_unref (img);
+	g_object_unref (pixbuf);
+}
+
+void GLView::RenderToCairo (cairo_t *cr, unsigned width, unsigned height) const
+{
+	double scale = 300. / 72.;
+	GdkPixbuf *pixbuf = BuildPixbuf (width * scale, height * scale);
+	GOImage *img = go_image_new_from_pixbuf (pixbuf);
+	cairo_pattern_t *cr_pattern = go_image_create_cairo_pattern (img);
+	cairo_matrix_t cr_matrix;
+	cairo_matrix_init_scale (&cr_matrix, scale, scale);
+	cairo_pattern_set_matrix (cr_pattern, &cr_matrix);
+	cairo_rectangle (cr, 0., 0., width, height);
+	cairo_set_source (cr, cr_pattern);
+	cairo_fill (cr);
+	cairo_pattern_destroy (cr_pattern);
+	g_object_unref (img);
+	g_object_unref (pixbuf);
+}
+
 }	//	namespace gcu
